"""Nonparametric rank correlation estimators as alternative to
 linear correlation estimators."""
from __future__ import absolute_import

import numpy as np
from scipy.stats import rankdata, kendalltau, weightedtau


def _compute_ranks(X, winsorize=False, truncation=None, verbose=True):
    """
    Transform each column into ranked data. Tied ranks are averaged.
    Ranks can optionally be winsorized as described in Liu 2009 otherwise
    this returns Tsukahara's scaled rank based Z-estimator.

    Parameters
    ----------
    X : array-like, shape = (n_samples, n_features)
        The data matrix where each column is a feature.
         Row observations for each column will be replaced
         by correponding rank.

    winsorize: bool
        Choose whether ranks should be winsorized (trimmed) or not. If True,
        then ranks will be winsorized using the truncation parameter.

    truncation: (float)
        The default value is given by 1/(4 n^(1/4) * sqrt(pi log n)), where
        n is the number of samples.

    Returns
    -------
    Xrank

    References
    ----------

    Liu, Han, John Lafferty, and Larry Wasserman.
    "The nonparanormal: Semiparametric estimation of high dimensional
    undirected graphs."
    Journal of Machine Learning Research 10.Oct (2009): 2295-2328.
    """
    n_samples, n_features = X.shape
    Xrank = np.zeros(shape=X.shape)

    if winsorize:
        if truncation is None:
            truncation = 1 / (
                4 * np.power(n_samples, 0.25) * np.sqrt(np.pi * np.log(n_samples))
            )

        elif truncation > 1:
            truncation = np.min(1.0, truncation)

    for col in np.arange(n_features):
        Xrank[:, col] = rankdata(X[:, col], method="average")
        Xrank[:, col] /= n_samples
        if winsorize:
            if n_samples > 100 * n_features:
                Xrank[:, col] = n_samples * Xrank[:, col] / (n_samples + 1)
            else:
                lower_truncate = Xrank[:, col] <= truncation
                upper_truncate = Xrank[:, col] > 1 - truncation
                Xrank[lower_truncate, col] = truncation
                Xrank[upper_truncate, col] = 1 - truncation

    return Xrank


def spearman_correlation(X, rowvar=False):
    """
    Computes the spearman correlation estimate.
    This is effectively a bias corrected pearson correlation
    between rank transformed columns of X.

    Parameters
    ----------
    X: array-like, shape = [n_samples, n_features]
        Data matrix using which we compute the empirical
        correlation

    Returns
    -------
    rank_correlation

    References
    ----------

    Xue, Lingzhou; Zou, Hui.
    "Regularized rank-based estimation of high-dimensional
    nonparanormal graphical models."
    Ann. Statist. 40 (2012), no. 5, 2541--2571. doi:10.1214/12-AOS1041.

    Liu, Han, Fang; Yuan, Ming; Lafferty, John; Wasserman, Larry.
    "High-dimensional semiparametric Gaussian copula graphical models."
    Ann. Statist. 40.4 (2012): 2293-2326. doi:10.1214/12-AOS1037
    """

    Xrank = _compute_ranks(X)
    rank_correlation = np.corrcoef(Xrank, rowvar=rowvar)

    return 2 * np.sin(rank_correlation * np.pi / 6)


def kendalltau_correlation(X, rowvar=False, weighted=False):
    """
    Computes kendall's tau correlation estimate.
    The option to use scipy.stats.weightedtau is not recommended
    as the implementation does not appear to handle ties correctly.

    Parameters
    ----------
    X: array-like, shape = [n_samples, n_features]
        Data matrix using which we compute the empirical
        correlation

    Returns
    -------
    rank_correlation

    References
    ----------

    Liu, Han, Fang; Yuan, Ming; Lafferty, John; Wasserman, Larry.
    "High-dimensional semiparametric Gaussian copula graphical models."
    Ann. Statist. 40.4 (2012): 2293-2326. doi:10.1214/12-AOS1037

    Barber, Rina Foygel; Kolar, Mladen.
    "ROCKET: Robust Confidence Intervals via Kendall's Tau
    for Transelliptical Graphical Models."
     arXiv:1502.07641
    """

    if rowvar:
        X = X.T

    _, n_features = X.shape
    rank_correlation = np.eye(n_features)
    for row in np.arange(n_features):
        for col in np.arange(1 + row, n_features):
            if weighted:
                rank_correlation[row, col], _ = weightedtau(
                    X[:, row], X[:, col], rank=False
                )
            else:
                rank_correlation[row, col], _ = kendalltau(X[:, row], X[:, col])
    rank_correlation = np.triu(rank_correlation, 1) + rank_correlation.T

    return np.sin(rank_correlation * np.pi / 2)


def winsorized_rank_correlation(X, rowvar=False, weighted=False):
    """
    Computes rank correlations using a winsorized ranks.
    Relevant to high dimensional settings where the n_samples < n_features
    resulting too large variance in the ranks.

    Parameters
    ----------
    X: array-like, shape = [n_samples, n_features]
        Data matrix using which we compute the empirical
        correlation

    Returns
    -------
    rank_correlation

    References
    ----------

    Liu, Han, John Lafferty, and Larry Wasserman.
    "The nonparanormal: Semiparametric estimation of high dimensional
    undirected graphs."
    Journal of Machine Learning Research 10.Oct (2009): 2295-2328.
    """
    pass



import numpy as np
cimport numpy as np
from libc.stdint cimport uint32_t, int32_t

cdef extern from "QUIC.h":
    void QUIC(char mode, uint32_t& p, double* S, double* Lambda0,
	  uint32_t& pathLen, double* path, double& tol,
	  int32_t& msg, uint32_t& maxIter,
	  double* X, double* W, double* opt, double* cputime,
	  uint32_t* iter, double* dGap)

def quic(char* mode, int p, 
        np.ndarray[np.float64_t, ndim=2, mode='c'] S, 
        np.ndarray[np.float64_t, ndim=2, mode='c'] L,
        int pathLen,
        np.ndarray[np.float64_t, ndim=1, mode='c'] path,
        double tol, int msg, int max_iter, 
        np.ndarray[np.float64_t, ndim=2, mode='c'] X, 
        np.ndarray[np.float64_t, ndim=2, mode='c'] W,
        np.ndarray[np.float64_t, ndim=1, mode='c'] opt,
        np.ndarray[np.float64_t, ndim=1, mode='c'] cputime,
        np.ndarray[np.uint32_t, ndim=1, mode='c'] iter,
        np.ndarray[np.float64_t, ndim=1, mode='c'] dGap
        ):
    """
    """
    cdef uint32_t _p = p
    cdef uint32_t _pathLen = pathLen
    cdef int32_t _msg = msg
    cdef uint32_t _max_iter = max_iter
    cdef uint32_t* _iter = <uint32_t*> iter.data
    QUIC(mode[0], _p, &S[0,0], &L[0,0], _pathLen, &path[0], tol, _msg, _max_iter, 
            &X[0,0], &W[0,0], &opt[0], &cputime[0], &_iter[0], &dGap[0])

    return



from __future__ import absolute_import

import sys
import time
import operator
import numpy as np
from functools import partial

from sklearn.covariance import EmpiricalCovariance
from sklearn.utils import check_array, as_float_array, deprecated
from numpy.testing import assert_array_almost_equal
from joblib import Parallel, delayed
from sklearn.model_selection import cross_val_score, RepeatedKFold

from . import pyquic
from .inverse_covariance import (
    InverseCovarianceEstimator,
    _init_coefs,
    _compute_error,
    _validate_path,
)


def quic(
    S,
    lam,
    mode="default",
    tol=1e-6,
    max_iter=1000,
    Theta0=None,
    Sigma0=None,
    path=None,
    msg=0,
):
    """Fits the inverse covariance model according to the given training
    data and parameters.

    Parameters
    -----------
    S : 2D ndarray, shape (n_features, n_features)
        Empirical covariance or correlation matrix.

    Other parameters described in `class InverseCovariance`.

    Returns
    -------
    Theta :
    Sigma :
    opt :
    cputime :
    iters :
    dGap :
    """
    assert mode in ["default", "path", "trace"], "mode = 'default', 'path' or 'trace'."

    Sn, Sm = S.shape
    if Sn != Sm:
        raise ValueError("Input data must be square. S shape = {}".format(S.shape))
        return

    # Regularization parameter matrix L.
    if isinstance(lam, float):
        _lam = np.empty((Sn, Sm))
        _lam[:] = lam
        _lam[np.diag_indices(Sn)] = 0.0  # make sure diagonal is zero
    else:
        assert lam.shape == S.shape, "lam, S shape mismatch."
        _lam = as_float_array(lam, copy=False, force_all_finite=False)

    # Defaults.
    optSize = 1
    iterSize = 1
    if mode is "trace":
        optSize = max_iter

    # Default Theta0, Sigma0 when both are None.
    if Theta0 is None and Sigma0 is None:
        Theta0 = np.eye(Sn)
        Sigma0 = np.eye(Sn)

    assert Theta0 is not None, "Theta0 and Sigma0 must both be None or both specified."
    assert Sigma0 is not None, "Theta0 and Sigma0 must both be None or both specified."
    assert Theta0.shape == S.shape, "Theta0, S shape mismatch."
    assert Sigma0.shape == S.shape, "Theta0, Sigma0 shape mismatch."
    Theta0 = as_float_array(Theta0, copy=False, force_all_finite=False)
    Sigma0 = as_float_array(Sigma0, copy=False, force_all_finite=False)

    if mode == "path":
        assert path is not None, "Please specify the path scaling values."

        # path must be sorted from largest to smallest and have unique values
        check_path = sorted(set(path), reverse=True)
        assert_array_almost_equal(check_path, path)

        path_len = len(path)
        optSize = path_len
        iterSize = path_len

        # Note here: memory layout is important:
        # a row of X/W holds a flattened Sn x Sn matrix,
        # one row for every element in _path_.
        Theta = np.empty((path_len, Sn * Sn))
        Theta[0, :] = Theta0.ravel()
        Sigma = np.empty((path_len, Sn * Sn))
        Sigma[0, :] = Sigma0.ravel()
    else:
        path = np.empty(1)
        path_len = len(path)

        Theta = np.empty(Theta0.shape)
        Theta[:] = Theta0
        Sigma = np.empty(Sigma0.shape)
        Sigma[:] = Sigma0

    # Cython fix for Python3
    # http://cython.readthedocs.io/en/latest/src/tutorial/strings.html
    quic_mode = mode
    if sys.version_info[0] >= 3:
        quic_mode = quic_mode.encode("utf-8")

    # Run QUIC.
    opt = np.zeros(optSize)
    cputime = np.zeros(optSize)
    dGap = np.zeros(optSize)
    iters = np.zeros(iterSize, dtype=np.uint32)
    pyquic.quic(
        quic_mode,
        Sn,
        S,
        _lam,
        path_len,
        path,
        tol,
        msg,
        max_iter,
        Theta,
        Sigma,
        opt,
        cputime,
        iters,
        dGap,
    )

    if optSize == 1:
        opt = opt[0]
        cputime = cputime[0]
        dGap = dGap[0]

    if iterSize == 1:
        iters = iters[0]

    # reshape Theta, Sigma in path mode
    Theta_out = Theta
    Sigma_out = Sigma
    if mode == "path":
        Theta_out = []
        Sigma_out = []
        for lidx in range(path_len):
            Theta_out.append(np.reshape(Theta[lidx, :], (Sn, Sn)))
            Sigma_out.append(np.reshape(Sigma[lidx, :], (Sn, Sn)))

    return Theta_out, Sigma_out, opt, cputime, iters, dGap


class QuicGraphicalLasso(InverseCovarianceEstimator):
    """
    Computes a sparse inverse covariance matrix estimation using quadratic
    approximation.

    The inverse covariance is estimated the sample covariance estimate
    $S$ as an input such that:

    $T_hat = max_{\Theta} logdet(Theta) - Trace(ThetaS) - \lambda|\Theta|_1 $

    Parameters
    -----------
    lam : scalar or 2D ndarray, shape (n_features, n_features) (default=0.5)
        Regularization parameters per element of the inverse covariance matrix.

        If a scalar lambda is used, a penalty matrix will be generated
        containing lambda for all values in both upper and lower triangles
        and zeros along the diagonal.  This differs from the scalar graphical
        lasso by the diagonal. To replicate the scalar formulation you must
        manualy pass in lam * np.ones((n_features, n_features)).

    mode : one of 'default', 'path', or 'trace'
        Computation mode.

    tol : float (default=1e-6)
        Convergence threshold.

    max_iter : int (default=1000)
        Maximum number of Newton iterations.

    Theta0 : 2D ndarray, shape (n_features, n_features) (default=None)
        Initial guess for the inverse covariance matrix. If not provided, the
        diagonal identity matrix is used.

    Sigma0 : 2D ndarray, shape (n_features, n_features) (default=None)
        Initial guess for the covariance matrix. If not provided the diagonal
        identity matrix is used.

    path : array of floats (default=None)
        In "path" mode, an array of float values for scaling lam.
        The path must be sorted largest to smallest.  This class will auto sort
        this, in which case indices correspond to self.path_

    method : 'quic' or 'bigquic', ... (default=quic)
        Currently only 'quic' is supported.

    verbose : integer
        Used in quic routine.

    score_metric : one of 'log_likelihood' (default), 'frobenius', 'spectral',
                  'kl', or 'quadratic'
        Used for computing self.score().

    init_method : one of 'corrcoef', 'cov', 'spearman', 'kendalltau',
        or a custom function.
        Computes initial covariance and scales lambda appropriately.
        Using the custom function extends graphical model estimation to
        distributions beyond the multivariate Gaussian.
        The `spearman` or `kendalltau` options extend inverse covariance
        estimation to nonparanormal and transelliptic graphical models.
        Custom function must return ((n_features, n_features) ndarray, float)
        where the scalar parameter will be used to scale the penalty lam.

    auto_scale : bool
        If True, will compute self.lam_scale_ = max off-diagonal value when
        init_method='cov'.
        If false, then self.lam_scale_ = 1.
        lam_scale_ is used to scale user-supplied self.lam during fit.

    Methods
    ----------
    lam_at_index(lidx) :  Compute the scaled lambda used at index lidx.
        The parameter lidx is ignored when mode='default'.  Can use self.lam_
        for convenience in this case.

    Attributes
    ----------
    covariance_ : 2D ndarray, shape (n_features, n_features)
        Estimated covariance matrix
        If mode='path', this is 2D ndarray, shape (len(path), n_features ** 2)

    precision_ : 2D ndarray, shape (n_features, n_features)
        Estimated pseudo-inverse matrix.
        If mode='path', this is 2D ndarray, shape (len(path), n_features ** 2)

    sample_covariance_ : 2D ndarray, shape (n_features, n_features)
        Estimated sample covariance matrix

    lam_ : (float) or 2D ndarray, shape (n_features, n_features)
        When mode='default', this is the lambda used in fit (lam * lam_scale_)

    lam_scale_ : (float)
        Additional scaling factor on lambda (due to magnitude of
        sample_covariance_ values).

    path_ : None or array of floats
        Sorted (largest to smallest) path.  This will be None if not in path
        mode.

    opt_ :

    cputime_ :

    iters_ :

    duality_gap_ :

    """

    def __init__(
        self,
        lam=0.5,
        mode="default",
        tol=1e-6,
        max_iter=1000,
        Theta0=None,
        Sigma0=None,
        path=None,
        method="quic",
        verbose=0,
        score_metric="log_likelihood",
        init_method="corrcoef",
        auto_scale=True,
    ):
        # quic-specific params
        self.lam = lam
        self.mode = mode
        self.tol = tol
        self.max_iter = max_iter
        self.Theta0 = Theta0
        self.Sigma0 = Sigma0
        self.method = method
        self.verbose = verbose
        self.path = path

        if self.mode == "path" and path is None:
            raise ValueError("path required in path mode.")
            return

        super(QuicGraphicalLasso, self).__init__(
            score_metric=score_metric, init_method=init_method, auto_scale=auto_scale
        )

    def fit(self, X, y=None, **fit_params):
        """Fits the inverse covariance model according to the given training
        data and parameters.

        Parameters
        -----------
        X : 2D ndarray, shape (n_features, n_features)
            Input data.

        Returns
        -------
        self
        """

        # To satisfy sklearn
        if y is not None and len(y) == 1:
            raise ValueError("Cannot fit with just 1 sample.")

        # quic-specific outputs
        self.opt_ = None
        self.cputime_ = None
        self.iters_ = None
        self.duality_gap_ = None

        # these must be updated upon self.fit()
        self.sample_covariance_ = None
        self.lam_scale_ = None
        self.is_fitted_ = False

        self.path_ = _validate_path(self.path)
        X = check_array(X, ensure_min_features=2, estimator=self)
        X = as_float_array(X, copy=False, force_all_finite=False)
        self.init_coefs(X)
        if self.method == "quic":
            (
                self.precision_,
                self.covariance_,
                self.opt_,
                self.cputime_,
                self.iters_,
                self.duality_gap_,
            ) = quic(
                self.sample_covariance_,
                self.lam * self.lam_scale_,
                mode=self.mode,
                tol=self.tol,
                max_iter=self.max_iter,
                Theta0=self.Theta0,
                Sigma0=self.Sigma0,
                path=self.path_,
                msg=self.verbose,
            )
        else:
            raise NotImplementedError("Only method='quic' has been implemented.")

        self.is_fitted_ = True
        self.n_features_in_ = X.shape[1]
        return self

    def lam_at_index(self, lidx):
        """Compute the scaled lambda used at index lidx."""
        if self.path_ is None:
            return self.lam * self.lam_scale_

        return self.lam * self.lam_scale_ * self.path_[lidx]

    @property
    def lam_(self):
        if self.path_ is not None:
            print("lam_ is an invalid parameter in path mode, " "use self.lam_at_index")
        return self.lam_at_index(0)


@deprecated(
    "The class QuicGraphLasso is deprecated " "Use class QuicGraphicalLasso instead."
)
class QuicGraphLasso(QuicGraphicalLasso):
    pass


def _quic_path(
    X,
    path,
    X_test=None,
    lam=0.5,
    tol=1e-6,
    max_iter=1000,
    Theta0=None,
    Sigma0=None,
    method="quic",
    verbose=0,
    score_metric="log_likelihood",
    init_method="corrcoef",
):
    """Wrapper to compute path for example X."""
    S, lam_scale_ = _init_coefs(X, method=init_method)

    path = path.copy(order="C")

    if method == "quic":
        (precisions_, covariances_, opt_, cputime_, iters_, duality_gap_) = quic(
            S,
            lam,
            mode="path",
            tol=tol,
            max_iter=max_iter,
            Theta0=Theta0,
            Sigma0=Sigma0,
            path=path,
            msg=verbose,
        )
    else:
        raise NotImplementedError("Only method='quic' has been implemented.")

    if X_test is not None:
        S_test, lam_scale_test = _init_coefs(X_test, method=init_method)

        path_errors = []
        for lidx, lam in enumerate(path):
            path_errors.append(
                _compute_error(
                    S_test,
                    covariances_[lidx],
                    precisions_[lidx],
                    score_metric=score_metric,
                )
            )
        scores_ = [-e for e in path_errors]

        return covariances_, precisions_, scores_

    return covariances_, precisions_


def _quic_path_spark(indexed_params, quic_path, X_bc):
    index, (local_train, local_test) = indexed_params
    result = quic_path(X_bc.value[local_train], X_test=X_bc.value[local_test])
    return index, result


class QuicGraphicalLassoCV(InverseCovarianceEstimator):
    """Sparse inverse covariance w/ cross-validated choice of the l1 penalty
    via quadratic approximation.

    This takes advantage of "path" mode in QuicGraphicalLasso.
    See sklearn.covariance.graph_lasso.GraphLassoCV.

    Parameters
    -----------
    lam : scalar or 2D ndarray, shape (n_features, n_features) (default=0.5)
        Regularization parameters per element of the inverse covariance matrix.
        The next parameter 'lams' scale this matrix as the lasso path learned.

        If a scalar lambda is used, a penalty matrix will be generated
        containing lambda for all values in both upper and lower triangles
        and zeros along the diagonal.  This differs from the scalar graphical
        lasso by the diagonal. To replicate the scalar formulation you must
        manualy pass in lam * np.ones((n_features, n_features)).

    lams : integer, or list positive float, optional
        If an integer is given, it fixes the number of points on the
        grids of alpha to be used. If a list is given, it gives the
        grid to be used. See the notes in the class docstring for
        more details.

    cv : int, cross-validation generator or an iterable, optional
        Determines the cross-validation splitting strategy.
        Possible inputs for cv are:

        - None, use the default 3-fold, 3-trial RepeatedKFold cross-validation,
        - integer, to specify the number of folds (3-trial RepeatedKFold).
        - tuple, (n_folds, n_trials)
        - An object to be used as a cross-validation generator.
        - An iterable yielding train/test splits.

    n_refinements : strictly positive integer
        The number of times the grid is refined. Not used if explicit
        values of alphas are passed.

    n_jobs : int, optional
        number of jobs to run in parallel (default 1).

    sc : sparkContext (default=None)
        If not None and a valid SparkContext, the cross-validation iterations
        will be performed in parallel via Apache Spark.  In this case n_jobs
        will be unused.

    tol : float (default=1e-6)
        Convergence threshold.

    max_iter : int (default=1000)
        Maximum number of Newton iterations.

    Theta0 : 2D ndarray, shape (n_features, n_features) (default=None)
        Initial guess for the inverse covariance matrix. If not provided, the
        diagonal identity matrix is used.

    Sigma0 : 2D ndarray, shape (n_features, n_features) (default=None)
        Initial guess for the covariance matrix. If not provided the diagonal
        identity matrix is used.

    method : 'quic' or 'bigquic', ... (default=quic)
        Currently only 'quic' is supported.

    score_metric : one of 'log_likelihood' (default), 'frobenius', 'spectral',
                  'kl', or 'quadratic'
        Used for computing self.score().

    init_method : one of 'corrcoef', 'cov', 'spearman', 'kendalltau',
        or a custom function.
        Computes initial covariance and scales lambda appropriately.
        Using the custom function extends graphical model estimation to
        distributions beyond the multivariate Gaussian.
        The `spearman` or `kendalltau` options extend inverse covariance
        estimation to nonparanormal and transelliptic graphical models.
        Custom function must return ((n_features, n_features) ndarray, float)
        where the scalar parameter will be used to scale the penalty lam.

    auto_scale : bool
        If True, will compute self.lam_scale_ = max off-diagonal value when
        init_method='cov'.
        If false, then self.lam_scale_ = 1.
        lam_scale_ is used to scale user-supplied self.lam during fit.

    backend : string, optional (default=threading)
        Joblib parallelization backend.
        Not used when using the sparkContext (sc).

    Attributes
    ----------
    covariance_ : 2D ndarray, shape (n_features, n_features)
        Estimated covariance matrix

    precision_ : 2D ndarray, shape (n_features, n_features)
        Estimated pseudo-inverse matrix.

    lam_ : float
        Penalization parameter selected.

    cv_lams_ : list of float
        All penalization parameters explored.

    grid_scores_: 2D numpy.ndarray (n_alphas, n_folds)
        Log-likelihood score on left-out data across folds.

    n_iter_ : int
        Number of iterations run for the optimal alpha.
    """

    def __init__(
        self,
        lam=1.0,
        lams=4,
        n_refinements=4,
        cv=None,
        tol=1e-6,
        max_iter=1000,
        Theta0=None,
        Sigma0=None,
        method="quic",
        verbose=0,
        n_jobs=1,
        sc=None,
        score_metric="log_likelihood",
        init_method="corrcoef",
        auto_scale=True,
        backend="threading",
    ):
        # GridCV params
        self.n_jobs = n_jobs
        self.sc = sc
        self.cv = cv
        self.lam = lam
        self.lams = lams
        self.n_refinements = n_refinements

        # quic-specific params
        self.tol = tol
        self.max_iter = max_iter
        self.Theta0 = Theta0
        self.Sigma0 = Sigma0
        self.method = method
        self.verbose = verbose
        self.backend = backend

        super(QuicGraphicalLassoCV, self).__init__(
            score_metric=score_metric, init_method=init_method, auto_scale=auto_scale
        )

    def fit(self, X, y=None):
        """Fits the GraphLasso covariance model to X.

        Closely follows sklearn.covariance.graph_lasso.GraphLassoCV.

        Parameters
        ----------
        X : ndarray, shape (n_samples, n_features)
            Data from which to compute the covariance estimate
        """

        # To satisfy sklearn
        if y is not None and len(y) == 1:
            raise ValueError("Cannot fit with just 1 sample.")

        # quic-specific outputs
        self.opt_ = None
        self.cputime_ = None
        self.iters_ = None
        self.duality_gap_ = None

        # these must be updated upon self.fit()
        self.sample_covariance_ = None
        self.lam_scale_ = None
        self.is_fitted_ = False

        # initialize
        X = check_array(X, ensure_min_features=2, estimator=self)
        X = as_float_array(X, copy=False, force_all_finite=False)

        if self.cv is None:
            cv = (3, 10)
        elif isinstance(self.cv, int):
            cv = (self.cv, 10)  # upgrade with default number of trials
        elif isinstance(self.cv, tuple):
            cv = self.cv

        cv = RepeatedKFold(n_splits=cv[0], n_repeats=cv[1])

        self.init_coefs(X)

        # get path
        if isinstance(self.lams, int):
            n_refinements = self.n_refinements
            lam_1 = self.lam_scale_
            lam_0 = 1e-2 * lam_1
            path = np.logspace(np.log10(lam_0), np.log10(lam_1), self.lams)[::-1]
        else:
            path = self.lams
            n_refinements = 1

        # run this thing a bunch
        results = list()
        t0 = time.time()
        for rr in range(n_refinements):
            if self.sc is None:
                # parallel version
                this_result = Parallel(
                    n_jobs=self.n_jobs, verbose=self.verbose, backend=self.backend
                )(
                    delayed(_quic_path)(
                        X[train],
                        path,
                        X_test=X[test],
                        lam=self.lam,
                        tol=self.tol,
                        max_iter=self.max_iter,
                        Theta0=self.Theta0,
                        Sigma0=self.Sigma0,
                        method=self.method,
                        verbose=self.verbose,
                        score_metric=self.score_metric,
                        init_method=self.init_method,
                    )
                    for train, test in cv.split(X)
                )
            else:
                # parallel via spark
                train_test_grid = [(train, test) for (train, test) in cv.split(X)]
                indexed_param_grid = list(
                    zip(range(len(train_test_grid)), train_test_grid)
                )
                par_param_grid = self.sc.parallelize(indexed_param_grid)
                X_bc = self.sc.broadcast(X)

                # wrap function parameters so we dont pick whole self object
                quic_path = partial(
                    _quic_path,
                    path=path,
                    lam=self.lam,
                    tol=self.tol,
                    max_iter=self.max_iter,
                    Theta0=self.Theta0,
                    Sigma0=self.Sigma0,
                    method=self.method,
                    verbose=self.verbose,
                    score_metric=self.score_metric,
                    init_method=self.init_method,
                )

                indexed_results = dict(
                    par_param_grid.map(
                        partial(_quic_path_spark, quic_path=quic_path, X_bc=X_bc)
                    ).collect()
                )
                this_result = [
                    indexed_results[idx] for idx in range(len(train_test_grid))
                ]
                X_bc.unpersist()

            # Little dance to transform the list in what we need
            covs, _, scores = zip(*this_result)
            covs = zip(*covs)
            scores = zip(*scores)
            results.extend(zip(path, scores, covs))
            results = sorted(results, key=operator.itemgetter(0), reverse=True)

            # Find the maximum (avoid using built in 'max' function to
            # have a fully-reproducible selection of the smallest alpha
            # in case of equality)
            best_score = -np.inf
            last_finite_idx = 0
            best_index = 0
            for index, (lam, scores, _) in enumerate(results):
                # sometimes we get -np.inf in the result (in kl-loss)
                scores = [s for s in scores if not np.isinf(s)]
                if len(scores) == 0:
                    this_score = -np.inf
                else:
                    this_score = np.mean(scores)

                if this_score >= 0.1 / np.finfo(np.float64).eps:
                    this_score = np.nan

                if np.isfinite(this_score):
                    last_finite_idx = index

                if this_score >= best_score:
                    best_score = this_score
                    best_index = index

            # Refine the grid
            if best_index == 0:
                # We do not need to go back: we have chosen
                # the highest value of lambda for which there are
                # non-zero coefficients
                lam_1 = results[0][0]
                lam_0 = results[1][0]

            elif best_index == last_finite_idx and not best_index == len(results) - 1:
                # We have non-converged models on the upper bound of the
                # grid, we need to refine the grid there
                lam_1 = results[best_index][0]
                lam_0 = results[best_index + 1][0]

            elif best_index == len(results) - 1:
                lam_1 = results[best_index][0]
                lam_0 = 0.01 * results[best_index][0]

            else:
                lam_1 = results[best_index - 1][0]
                lam_0 = results[best_index + 1][0]

            if isinstance(self.lams, int):
                path = np.logspace(np.log10(lam_1), np.log10(lam_0), self.lams + 2)
                path = path[1:-1]

            if self.verbose and n_refinements > 1:
                print(
                    "[GraphLassoCV] Done refinement % 2i out of %i: % 3is"
                    % (rr + 1, n_refinements, time.time() - t0)
                )

        results = list(zip(*results))
        grid_scores_ = list(results[1])
        lams = list(results[0])

        # Finally, compute the score with lambda = 0
        lams.append(0)
        grid_scores_.append(
            cross_val_score(EmpiricalCovariance(), X, cv=cv, n_jobs=self.n_jobs)
        )
        self.grid_scores_ = np.array(grid_scores_)
        self.lam_ = self.lam * lams[best_index]
        self.cv_lams_ = [self.lam * l for l in lams]

        # Finally fit the model with the selected lambda
        if self.method == "quic":
            (
                self.precision_,
                self.covariance_,
                self.opt_,
                self.cputime_,
                self.iters_,
                self.duality_gap_,
            ) = quic(
                self.sample_covariance_,
                self.lam_,
                mode="default",
                tol=self.tol,
                max_iter=self.max_iter,
                Theta0=self.Theta0,
                Sigma0=self.Sigma0,
                path=None,
                msg=self.verbose,
            )
        else:
            raise NotImplementedError("Only method='quic' has been implemented.")

        self.is_fitted_ = True
        self.n_features_in_ = X.shape[1]
        return self


@deprecated(
    "The class QuicGraphLassoCV is deprecated "
    "Use class QuicGraphicalLassoCV instead."
)
class QuicGraphLassoCV(QuicGraphicalLassoCV):
    pass


class QuicGraphicalLassoEBIC(InverseCovarianceEstimator):
    """
    Computes a sparse inverse covariance matrix estimation using quadratic
    approximation and EBIC model selection. (Convenience Class)

    Note: This estimate can be obtained using the more general QuicGraphicalLasso
          estimator and taking advantage of `ebic_select()` and
          `lambda_at_index()` methods.

    See analogous sklearn.linear_model.LassoLarsIC.

    Parameters
    -----------
    lam : scalar or 2D ndarray, shape (n_features, n_features) (default=0.5)
        Regularization parameters per element of the inverse covariance matrix.

        If a scalar lambda is used, a penalty matrix will be generated
        containing lambda for all values in both upper and lower triangles
        and zeros along the diagonal.  This differs from the scalar graphical
        lasso by the diagonal. To replicate the scalar formulation you must
        manualy pass in lam * np.ones((n_features, n_features)).

    path : array of floats or int (default=100)
        An array of float values for scaling lam.
        An int will choose the number of log-scale points to fit.

    gamma : float (default=0)
        Extended Bayesian Information Criteria (EBIC) for model selection.
        Choice of gamma=0 leads to classical BIC
        Positive gamma leads to stronger penalization of large graphs.

    tol : float (default=1e-6)
        Convergence threshold.

    max_iter : int (default=1000)
        Maximum number of Newton iterations.

    Theta0 : 2D ndarray, shape (n_features, n_features) (default=None)
        Initial guess for the inverse covariance matrix. If not provided, the
        diagonal identity matrix is used.

    Sigma0 : 2D ndarray, shape (n_features, n_features) (default=None)
        Initial guess for the covariance matrix. If not provided the diagonal
        identity matrix is used.

    method : one of 'quic'... (default=quic)

    verbose : integer
        Used in quic routine.

    init_method : one of 'corrcoef', 'cov', 'spearman', 'kendalltau',
        or a custom function.
        Computes initial covariance and scales lambda appropriately.
        Using the custom function extends graphical model estimation to
        distributions beyond the multivariate Gaussian.
        The `spearman` or `kendalltau` options extend inverse covariance
        estimation to nonparanormal and transelliptic graphical models.
        Custom function must return ((n_features, n_features) ndarray, float)
        where the scalar parameter will be used to scale the penalty lam.

    auto_scale : bool
        If True, will compute self.lam_scale_ = max off-diagonal value when
        init_method='cov'.
        If false, then self.lam_scale_ = 1.
        lam_scale_ is used to scale user-supplied self.lam during fit.

    Attributes
    ----------
    covariance_ : 2D ndarray, shape (n_features, n_features)
        Estimated covariance matrix

    precision_ : 2D ndarray, shape (n_features, n_features)
        Estimated pseudo-inverse matrix.

    sample_covariance_ : 2D ndarray, shape (n_features, n_features)
        Estimated sample covariance matrix

    lam_ : (float)
        Lambda chosen by EBIC (with scaling already applied).
    """

    def __init__(
        self,
        lam=1.0,
        path=100,
        gamma=0,
        tol=1e-6,
        max_iter=1000,
        Theta0=None,
        Sigma0=None,
        method="quic",
        verbose=0,
        score_metric="log_likelihood",
        init_method="corrcoef",
        auto_scale=True,
    ):
        # quic-specific params
        self.lam = lam
        self.tol = tol
        self.max_iter = max_iter
        self.Theta0 = Theta0
        self.Sigma0 = Sigma0
        self.method = method
        self.verbose = verbose
        self.path = path
        self.gamma = gamma

        super(QuicGraphicalLassoEBIC, self).__init__(
            init_method=init_method, score_metric=score_metric, auto_scale=auto_scale
        )

    def fit(self, X, y=None, **fit_params):
        """Fits the inverse covariance model according to the given training
        data and parameters.

        Parameters
        -----------
        X : 2D ndarray, shape (n_features, n_features)
            Input data.

        Returns
        -------
        self
        """

        # To satisfy sklearn
        if y is not None and len(y) == 1:
            raise ValueError("Cannot fit with just 1 sample.")

        # quic-specific outputs
        self.opt_ = None
        self.cputime_ = None
        self.iters_ = None
        self.duality_gap_ = None

        # these must be updated upon self.fit()
        self.path_ = None
        self.sample_covariance_ = None
        self.lam_scale_ = None
        self.lam_ = None
        self.is_fitted_ = False

        X = check_array(X, ensure_min_features=2, estimator=self)
        X = as_float_array(X, copy=False, force_all_finite=False)
        self.init_coefs(X)

        # either use passed in path, or make our own path
        lam_1 = self.lam_scale_
        lam_0 = 1e-2 * lam_1
        if self.path is None:
            self.path_ = np.logspace(np.log10(lam_0), np.log10(lam_1), 100)[::-1]
        elif isinstance(self.path, int):
            self.path_ = np.logspace(np.log10(lam_0), np.log10(lam_1), self.path)[::-1]
        else:
            self.path_ = self.path

        self.path_ = _validate_path(self.path_)

        # fit along the path, temporarily populate
        # self.precision_, self.covariance_ with path values so we can use our
        # inherited selection function
        if self.method == "quic":
            (self.precision_, self.covariance_, _, _, _, _) = quic(
                self.sample_covariance_,
                self.lam * self.lam_scale_,
                mode="path",
                tol=self.tol,
                max_iter=self.max_iter,
                Theta0=self.Theta0,
                Sigma0=self.Sigma0,
                path=self.path_,
                msg=self.verbose,
            )
            self.is_fitted_ = True
        else:
            raise NotImplementedError("Only method='quic' has been implemented.")

        # apply EBIC criteria
        best_lam_idx = self.ebic_select(gamma=self.gamma)
        self.lam_ = self.lam * self.lam_scale_ * self.path_[best_lam_idx]
        self.precision_ = self.precision_[best_lam_idx]
        self.covariance_ = self.covariance_[best_lam_idx]

        self.is_fitted_ = True
        self.n_features_in_ = X.shape[1]
        return self


@deprecated(
    "The class QuicGraphLassoEBIC is deprecated "
    "Use class QuicGraphicalLassoEBIC instead."
)
class QuicGraphLassoEBIC(QuicGraphicalLassoEBIC):
    pass



import numpy as np


def _nonzero_intersection(m, m_hat):
    """Count the number of nonzeros in and between m and m_hat.

    Returns
    ----------
    m_nnz :  number of nonzeros in m (w/o diagonal)

    m_hat_nnz : number of nonzeros in m_hat (w/o diagonal)

    intersection_nnz : number of nonzeros in intersection of m/m_hat
                      (w/o diagonal)
    """
    n_features, _ = m.shape

    m_no_diag = m.copy()
    m_no_diag[np.diag_indices(n_features)] = 0
    m_hat_no_diag = m_hat.copy()
    m_hat_no_diag[np.diag_indices(n_features)] = 0

    m_hat_nnz = len(np.nonzero(m_hat_no_diag.flat)[0])
    m_nnz = len(np.nonzero(m_no_diag.flat)[0])

    intersection_nnz = len(
        np.intersect1d(np.nonzero(m_no_diag.flat)[0], np.nonzero(m_hat_no_diag.flat)[0])
    )

    return m_nnz, m_hat_nnz, intersection_nnz


def support_false_positive_count(m, m_hat):
    """Count the number of false positive support elements in
    m_hat in one triangle, not including the diagonal.
    """
    m_nnz, m_hat_nnz, intersection_nnz = _nonzero_intersection(m, m_hat)
    return int((m_hat_nnz - intersection_nnz) / 2.0)


def support_false_negative_count(m, m_hat):
    """Count the number of false negative support elements in
    m_hat in one triangle, not including the diagonal.
    """
    m_nnz, m_hat_nnz, intersection_nnz = _nonzero_intersection(m, m_hat)
    return int((m_nnz - intersection_nnz) / 2.0)


def support_difference_count(m, m_hat):
    """Count the number of different elements in the support in one triangle,
    not including the diagonal.
    """
    m_nnz, m_hat_nnz, intersection_nnz = _nonzero_intersection(m, m_hat)
    return int((m_nnz + m_hat_nnz - (2 * intersection_nnz)) / 2.0)


def has_exact_support(m, m_hat):
    """Returns 1 if support_difference_count is zero, 0 else."""
    m_nnz, m_hat_nnz, intersection_nnz = _nonzero_intersection(m, m_hat)
    return int((m_nnz + m_hat_nnz - (2 * intersection_nnz)) == 0)


def has_approx_support(m, m_hat, prob=0.01):
    """Returns 1 if model selection error is less than or equal to prob rate,
    0 else.

    NOTE: why does np.nonzero/np.flatnonzero create so much problems?
    """
    m_nz = np.flatnonzero(np.triu(m, 1))
    m_hat_nz = np.flatnonzero(np.triu(m_hat, 1))

    upper_diagonal_mask = np.flatnonzero(np.triu(np.ones(m.shape), 1))
    not_m_nz = np.setdiff1d(upper_diagonal_mask, m_nz)

    intersection = np.in1d(m_hat_nz, m_nz)  # true positives
    not_intersection = np.in1d(m_hat_nz, not_m_nz)  # false positives

    true_positive_rate = 0.0
    if len(m_nz):
        true_positive_rate = 1.0 * np.sum(intersection) / len(m_nz)
        true_negative_rate = 1.0 - true_positive_rate

    false_positive_rate = 0.0
    if len(not_m_nz):
        false_positive_rate = 1.0 * np.sum(not_intersection) / len(not_m_nz)

    return int(np.less_equal(true_negative_rate + false_positive_rate, prob))


def error_fro(m, m_hat):
    return np.linalg.norm(np.triu(m - m_hat, 1), ord="fro")



from __future__ import absolute_import

import numpy as np
from sklearn.base import BaseEstimator, clone
from sklearn.utils import check_array, as_float_array
from joblib import Parallel, delayed
from functools import partial

from .inverse_covariance import _init_coefs
from . import QuicGraphicalLasso


def _check_psd(m):
    return np.all(np.linalg.eigvals(m) >= 0)


def _fully_random_weights(n_features, lam_scale, prng):
    """Generate a symmetric random matrix with zeros along the diagonal."""
    weights = np.zeros((n_features, n_features))
    n_off_diag = int((n_features**2 - n_features) / 2)
    weights[np.triu_indices(n_features, k=1)] = 0.1 * lam_scale * prng.randn(
        n_off_diag
    ) + (0.25 * lam_scale)
    weights[weights < 0] = 0
    weights = weights + weights.T
    return weights


def _random_weights(n_features, lam, lam_perturb, prng):
    """Generate a symmetric random matrix with zeros along the diagnoal and
    non-zero elements take the value {lam * lam_perturb, lam / lam_perturb}
    with probability 1/2.
    """
    weights = np.zeros((n_features, n_features))
    n_off_diag = int((n_features**2 - n_features) / 2)
    berns = prng.binomial(1, 0.5, size=n_off_diag)
    vals = np.zeros(berns.shape)
    vals[berns == 0] = 1.0 * lam * lam_perturb
    vals[berns == 1] = 1.0 * lam / lam_perturb
    weights[np.triu_indices(n_features, k=1)] = vals
    weights[weights < 0] = 0
    weights = weights + weights.T
    return weights


def _fix_weights(weight_fun, *args):
    """Ensure random weight matrix is valid.

    TODO:  The diagonally dominant tuning currently doesn't make sense.
           Our weight matrix has zeros along the diagonal, so multiplying by
           a diagonal matrix results in a zero-matrix.
    """
    weights = weight_fun(*args)

    # TODO: fix this
    # disable checks for now
    return weights

    # if positive semidefinite, then we're good as is
    if _check_psd(weights):
        return weights

    # make diagonally dominant
    off_diag_sums = np.sum(weights, axis=1)  # NOTE: assumes diag is zero
    mod_mat = np.linalg.inv(np.sqrt(np.diag(off_diag_sums)))
    return np.dot(mod_mat, weights, mod_mat)


def _default_bootstrap(n_samples, num_subsamples, prng):
    """Returns an array of integers (0, n_samples-1)^num_subsamples."""
    return prng.permutation(n_samples)[:num_subsamples]


def _fit(
    indexed_params,
    penalization,
    lam,
    lam_perturb,
    lam_scale_,
    estimator,
    penalty_name,
    subsample,
    bootstrap,
    prng,
    X=None,
):
    """Wrapper function outside of instance for fitting a single model average
    trial.

    If X is None, then we assume we are using a broadcast spark object. Else,
    we expect X to get passed into this function.
    """
    index = indexed_params

    if isinstance(X, np.ndarray):
        local_X = X
    else:
        local_X = X.value

    n_samples, n_features = local_X.shape

    prec_is_real = False
    while not prec_is_real:
        boot_lam = None
        if penalization == "subsampling":
            pass
        elif penalization == "random":
            boot_lam = _fix_weights(_random_weights, n_features, lam, lam_perturb, prng)
        elif penalization == "fully-random":
            boot_lam = _fix_weights(_fully_random_weights, n_features, lam_scale_, prng)
        else:
            raise NotImplementedError(
                (
                    "Only penalization = 'subsampling', "
                    "'random', and 'fully-random' have "
                    "been implemented. Found {}.".format(penalization)
                )
            )

        # new instance of estimator
        new_estimator = clone(estimator)
        if boot_lam is not None:
            new_estimator.set_params(**{penalty_name: boot_lam})

        # fit estimator
        num_subsamples = int(subsample * n_samples)
        rp = bootstrap(n_samples, num_subsamples, prng)
        new_estimator.fit(local_X[rp, :])

        # check that new_estimator.precision_ is real
        # if not, skip this boot_lam and try again
        if isinstance(new_estimator.precision_, list):
            prec_real_bools = []
            for prec in new_estimator.precision_:
                prec_real_bools.append(np.all(np.isreal(prec)))

            prec_is_real = np.all(np.array(prec_real_bools) is True)

        elif isinstance(new_estimator.precision_, np.ndarray):
            prec_is_real = np.all(np.isreal(new_estimator.precision_))

        else:
            raise ValueError("Estimator returned invalid precision_.")

    return index, (boot_lam, rp, new_estimator)


def _cpu_map(fun, param_grid, n_jobs, verbose=True):
    return Parallel(
        n_jobs=n_jobs,
        verbose=verbose,
        backend="threading",  # any sklearn backend should work here
    )(delayed(fun)(params) for params in param_grid)


def _spark_map(fun, indexed_param_grid, sc, seed, X_bc):
    """We cannot pass a RandomState instance to each spark worker since it will
    behave identically across partitions.  Instead, we explictly handle the
    partitions with a newly seeded instance.

    The seed for each partition will be the "seed" (MonteCarloProfile.seed) +
    "split_index" which is the partition index.

    Following this trick:
        https://wegetsignal.wordpress.com/2015/05/08/
                generating-random-numbers-for-rdd-in-spark/
    """

    def _wrap_random_state(split_index, partition):
        prng = np.random.RandomState(seed + split_index)
        yield map(partial(fun, prng=prng, X=X_bc), partition)

    par_param_grid = sc.parallelize(indexed_param_grid)
    indexed_results = par_param_grid.mapPartitionsWithIndex(
        _wrap_random_state
    ).collect()
    return [item for sublist in indexed_results for item in sublist]


class ModelAverage(BaseEstimator):
    """
    Randomized model averaging meta-estimator.

    See analogous sklearn.linear_model.BaseRandomizedLinearModel.

    Parameters
    -----------
    estimator : An inverse covariance estimator instance
        After being fit, estimator.precision_ must either be a matrix with the
        precision or a list of precision matrices (e.g., path mode).

    n_trials : int (default=100)
        Number of random subsets for which to bootstrap the data.

    subsample : float in range (0, 1) (default=0.3)
        Fraction of examples to subsample in each bootstrap trial.

    normalize : bool (default=True)
        Determines whether the proportion_ matrix should be normalized to have
        values in the range (0, 1) or should be absolute.

    lam : float (default=0.5)
        Scalar lambda penalty used in penalization='random' mode.  Will be
        ignored in all other modes.

    lam_perturb : float \in (0, 1) (default=0.5)
        Scalar perturbation parameter used in penalization='random'.  Will be
        ignored in all other modes.

    init_method : one of 'corrcoef', 'cov', 'spearman', 'kendalltau',
        or a custom function. See InverseCovarianceEstimator

    penalization : one of 'subsampling', 'random' (default), 'fully-random'
        Strategy for generating new random penalization in each trial.

        subsampling: Only the observations will be subsampled, the original
                     penalty supplied in the estimator instance will be used.
                     Use this technique when the estimator does not support
                     matrix penalization (e.g., sklearn GraphicalLasso).

        random: In addition to randomly subsampling the observations, 'random'
                applies a randomly-perturbed 'lam' weight matrix.  The entries
                of the matrix take the value
                {lam * lam_perturb, lam / lam_perturb} with probability 1/2.
                User must supply a scalar 'lam' and 'lam_perturb' parameters.

        fully-random: In addition to randomly subsampling the observations,
                      'fully-random' generates a symmetric Gaussian matrix
                      appropriately scaled for the data.

        For more information on 'random' penalization, see:
            "Stability Selection"
            N. Meinhausen and P. Buehlmann, May 2009

            "Random lasso"
            S. Wang, B. Nan, S. Rosset, and J. Zhu, Apr 2011

        For more information on 'fully-random', see:
            "Mixed effects models for resampled network statistics improves
            statistical power to find differences in multi-subject functional
            connectivity"
            M. Narayan and G. Allen, March 2016

    support_thresh : float (0, 1)
        Threshold for estimating supports from proportions.  This is provided
        for convience.

    penalty_name : string (default='lam')
        Name of the penalty kwarg in the estimator.  This parameter is
        unimportant if penalization='subsampling'.

    bootstrap : callable fun (default=_default_bootstrap)
        A function that takes n_samples, num_subsamples as inputs and returns
        a list of sample indices in the range (0, n_samples-1).
        By default, indices are uniformly subsampled.

    n_jobs: int (optional)
        number of jobs to run in parallel (default 1).

    sc: sparkContext (optional)
        If a sparkContext object is provided, n_jobs will be ignore and the
        work will be parallelized via spark.

    seed : np.random.RandomState starting seed. (default=2)

    Attributes
    ----------
    proportion_ : matrix of size (n_features, n_features)
        Each entry indicates the sample probability (or count) of whether the
        inverse covariance is non-zero.

    support_ : matrix of size (n_features, n_features)
        Support estimate via thresholding proportions by support_thresh.

    estimators_ : list of estimator instances (n_trials, )
        The estimator instance from each trial.
        This returns an empty list if use_cache=False.

    lams_ : list of penalization matrices (n_trials, )
        The penalization matrix chosen in each trial.
        This returns an empty list if penalization='subsampling'.

    subsets_ : list of subset indices (n_trials, )
        The example indices chosen in each trial.
        This returns an empty list if use_cache=False.

    lam_ : float
        Average matrix value used among lam_ for all estimators.
    """

    def __init__(
        self,
        estimator=None,
        n_trials=100,
        subsample=0.3,
        normalize=True,
        lam=0.5,
        lam_perturb=0.5,
        init_method="corrcoef",
        penalization="random",
        penalty_name="lam",
        support_thresh=0.5,
        bootstrap=_default_bootstrap,
        n_jobs=1,
        sc=None,
        seed=1,
    ):
        self.estimator = estimator
        self.n_trials = n_trials
        self.subsample = subsample
        self.normalize = normalize
        self.lam = lam
        self.lam_perturb = lam_perturb
        self.init_method = init_method
        self.penalization = penalization
        self.penalty_name = penalty_name
        self.support_thresh = support_thresh
        self.bootstrap = bootstrap
        self.n_jobs = n_jobs
        self.sc = sc
        self.seed = seed

    def fit(self, X, y=None):
        """Learn a model averaged proportion matrix for X.
        Parameters
        ----------
        X : ndarray, shape (n_samples, n_features)
            Data from which to compute the proportion matrix.
        """
        # default to QuicGraphicalLasso
        estimator = self.estimator or QuicGraphicalLasso()
        # To satisfy sklearn
        if y is not None and len(y) == 1:
            raise ValueError("Cannot fit with just 1 sample.")
        self._prng = np.random.RandomState(self.seed)
        if self.penalization != "subsampling" and not hasattr(
            estimator, self.penalty_name
        ):
            raise ValueError(
                (
                    "Must specify valid penalty for "
                    "estimator: {}.".format(self.penalty_name)
                )
            )

        self.proportion_ = None
        self.support_ = None
        self.lam_ = None
        self.lam_scale_ = None
        self.estimators_ = []
        self.lams_ = []
        self.subsets_ = []

        X = check_array(X, ensure_min_features=2, estimator=self)
        X = as_float_array(X, copy=False, force_all_finite=False)

        n_samples_, n_features_ = X.shape
        _, self.lam_scale_ = _init_coefs(X, method=self.init_method)

        fit_fun = partial(
            _fit,
            penalization=self.penalization,
            lam=self.lam,
            lam_perturb=self.lam_perturb,
            lam_scale_=self.lam_scale_,
            estimator=estimator,
            penalty_name=self.penalty_name,
            subsample=self.subsample,
            bootstrap=self.bootstrap,
        )
        indexed_param_grid = [(nn,) for nn in range(self.n_trials)]

        if self.sc is None:
            results = _cpu_map(
                partial(fit_fun, X=X, prng=self._prng),
                indexed_param_grid,
                n_jobs=self.n_jobs,
            )
        else:
            X_bc = self.sc.broadcast(X)
            results = _spark_map(fit_fun, indexed_param_grid, self.sc, self.seed, X_bc)
            X_bc.unpersist()

        self.estimators_ = [e for r, (l, s, e) in results]
        self.subsets_ = [s for r, (l, s, e) in results]
        self.lams_ = [l for r, (l, s, e) in results]

        # reduce
        self.lam_ = 0.0
        self.proportion_ = np.zeros((n_features_, n_features_))
        for new_estimator in self.estimators_:
            # update proportions
            if isinstance(new_estimator.precision_, list):
                for prec in new_estimator.precision_:
                    self.proportion_[np.nonzero(prec)] += 1.0

            elif isinstance(new_estimator.precision_, np.ndarray):
                self.proportion_[np.nonzero(new_estimator.precision_)] += 1.0

            else:
                raise ValueError("Estimator returned invalid precision_.")

            # currently, dont estimate self.lam_ if penalty_name is different
            if self.penalty_name == "lam":
                self.lam_ += np.mean(new_estimator.lam_.flat)

        # estimate support locations
        threshold = self.support_thresh * self.n_trials
        self.support_ = np.zeros(self.proportion_.shape)
        self.support_[self.proportion_ > threshold] = 1.0

        self.lam_ /= self.n_trials
        if self.normalize:
            self.proportion_ /= self.n_trials
        self.n_features_in_ = X.shape[1]
        return self

    @property
    def precision_(self):
        """Convenience property to make compatible with AdaptiveGraphicalLasso.
        This is not a very good precision estimate.
        """
        return self.support_

    @property
    def covariance_(self):
        """Convenience property to make compatible with AdaptiveGraphicalLasso.
        This is not a very good covariance estimate.
        """
        return np.linalg.inv(self.support_)



from __future__ import absolute_import
from .monte_carlo_profile import MonteCarloProfile
from .metrics import (
    support_false_positive_count,
    support_false_negative_count,
    support_difference_count,
    has_exact_support,
    has_approx_support,
    error_fro,
)
from .graphs import lattice, blocks, Graph
from .cluster_graph import ClusterGraph
from .lattice_graph import LatticeGraph
from .erdos_renyi_graph import ErdosRenyiGraph


__all__ = [
    "MonteCarloProfile",
    "support_false_positive_count",
    "support_false_negative_count",
    "support_difference_count",
    "has_exact_support",
    "has_approx_support",
    "error_fro",
    "lattice",
    "blocks",
    "Graph",
    "ClusterGraph",
    "LatticeGraph",
    "ErdosRenyiGraph",
]



from __future__ import absolute_import

import numpy as np
from sklearn.datasets import make_sparse_spd_matrix
from .graphs import Graph


class ErdosRenyiGraph(Graph):
    """Returns the adjacency matrix for Erdos-Renyi network via .create().

    Parameters
    -----------
    spd_low : float (0, 1)
        Equivalent to make_sparse_spd_matrix `smallest_coef`

    spd_high : float (0, 1)
        Equivalent to make_sparse_spd_matrix `largest_coef`

    n_blocks : int (default=2)
        Number of blocks.  Returned matrix will be square with
        shape n_block_features * n_blocks.

    chain_blocks : bool (default=True)
        Apply random lattice structure to chain blocks together.

    seed : int
        Seed for np.random.RandomState seed. (default=1)
    """

    def __init__(self, spd_low=0.7, spd_high=0.7, **kwargs):
        self.spd_low = spd_low
        self.spd_high = spd_high
        super(ErdosRenyiGraph, self).__init__(**kwargs)

    def prototype_adjacency(self, n_block_features, alpha):
        """Build a new graph.

        Doc for ".create(n_features, alpha)"

        Parameters
        -----------
        n_features : int

        alpha : float (0,1)
            The complexity / sparsity factor.
            This is (1 - alpha_0) in sklearn.datasets.make_sparse_spd_matrix
            where alpha_0 is the probability that a coefficient is zero.

        Returns
        -----------
        (n_features, n_features) matrices: covariance, precision, adjacency
        """
        return make_sparse_spd_matrix(
            n_block_features,
            alpha=np.abs(1.0 - alpha),
            smallest_coef=self.spd_low,
            largest_coef=self.spd_high,
            random_state=self.prng,
        )



from __future__ import absolute_import

import numpy as np
from .graphs import Graph


class ClusterGraph(Graph):
    """Returns the adjacency matrix for a cluster network via .create().

    The graph can be made fully connected using chaining assumption when
    chain_blocks=True (default).

    Parameters
    -----------
    low : float (0, 1) (default=0.3)
        Lower bound for np.random.RandomState.uniform cluster values.

    high : float (0, 1) > low (default=0.7)
        Upper bound for np.random.RandomState.uniform vluster values.

    n_blocks : int (default=2)
        Number of blocks.  Returned matrix will be square with
        shape n_block_features * n_blocks.

    chain_blocks : bool (default=True)
        Apply random lattice structure to chain blocks together.

    seed : int
        Seed for np.random.RandomState seed. (default=1)
    """

    def __init__(self, low=0.7, high=0.7, **kwargs):
        self.low = low
        self.high = high
        super(ClusterGraph, self).__init__(**kwargs)

    def prototype_adjacency(self, n_block_features, alpha=None):
        """Build a new graph.

        Doc for ".create(n_features, alpha)"

        Parameters
        -----------
        n_features : int

        [alpha] : float (0,1)
            Unused.

        Each graph will have a minimum of

            (n_blocks * n_block_features**2 - n_blocks) / 2

        edges and exactly this amount if chain_blocks=False.

        Returns
        -----------
        (n_features, n_features) matrices: covariance, precision, adjacency
        """
        return -np.ones((n_block_features, n_block_features)) * 0.5 + self.prng.uniform(
            low=self.low, high=self.high, size=(n_block_features, n_block_features)
        )



from __future__ import absolute_import

import numpy as np
from functools import partial
from sklearn.base import clone
from joblib import Parallel, delayed

from .metrics import error_fro
from .erdos_renyi_graph import ErdosRenyiGraph
from .. import QuicGraphicalLasso, QuicGraphicalLassoCV


def _sample_mvn(n_samples, cov, prng):
    """Draw a multivariate normal sample from the graph defined by cov.

    Parameters
    -----------
    n_samples : int

    cov : matrix of shape (n_features, n_features)
        Covariance matrix of the graph.

    prng : np.random.RandomState instance.
    """
    n_features, _ = cov.shape
    return prng.multivariate_normal(np.zeros(n_features), cov, size=n_samples)


def _ms_fit(indexed_params, estimator, n_features, graph, prng):
    # unpack params
    index, (alpha, grid_point) = indexed_params

    # draw a new fixed graph for alpha
    cov, prec, adj = graph.create(n_features, alpha)

    # model selection (once per n_samples grid point)
    n_samples = int(grid_point * n_features)
    X = _sample_mvn(n_samples, cov, prng)
    ms_estimator = clone(estimator)
    ms_estimator.fit(X)

    return index, ((cov, prec, adj), ms_estimator.lam_, n_samples)


def _mc_fit(indexed_params, estimator, metrics, prng):
    # unpack params
    index, (nn, (cov, prec, adj), lam, n_samples) = indexed_params

    # compute mc trial
    X = _sample_mvn(n_samples, cov, prng)
    mc_estimator = clone(estimator)
    mc_estimator.set_params(lam=lam)
    mc_estimator.fit(X)
    results = {k: f(prec, mc_estimator.precision_) for k, f in metrics.items()}

    return index, results


def _cpu_map(fun, param_grid, n_jobs, verbose):
    return Parallel(
        n_jobs=n_jobs,
        verbose=verbose,
        backend="threading",  # any sklearn backend should work here
    )(delayed(fun)(params) for params in param_grid)


def _spark_map(fun, indexed_param_grid, sc, seed):
    """We cannot pass a RandomState instance to each spark worker since it will
    behave identically across partitions.  Instead, we explictly handle the
    partitions with a newly seeded instance.

    The seed for each partition will be the "seed" (MonteCarloProfile.seed) +
    "split_index" which is the partition index.

    Following this trick:
        https://wegetsignal.wordpress.com/2015/05/08/
                generating-random-numbers-for-rdd-in-spark/
    """

    def _wrap_random_state(split_index, partition):
        prng = np.random.RandomState(seed + split_index)
        yield map(partial(fun, prng=prng), partition)

    par_param_grid = sc.parallelize(indexed_param_grid)
    indexed_results = par_param_grid.mapPartitionsWithIndex(
        _wrap_random_state
    ).collect()
    return [item for sublist in indexed_results for item in sublist]


class MonteCarloProfile(object):
    """Compute performance metrics over multiple random trials (multivariate
    normal sample instances for a given graph).

    The graph is sampled once to obtain a penalty (via the ms_estimator) and
    then sampled and fit for n_trials for each (alpha, n_samples) parameter
    pair.

    Parameters
    -----------
    n_features : int (default=50)
        Fixed number of features to test.

    n_trials : int (default=100)
        Number of examples to draw to measure P(recovery).

    ms_estimator : "Model selection estimator" (default=None)
        An inverse covariance estimator instance. This estimator must be able
        to select a penalization parameter that can be accessed via the
        instance variable .lam_.

    mc_estimator : "Monte Carlo trial estimator" (default=None)
        An inverse covariance estimator instance. Estimator to use on each
        instance after selecting a penalty lambda. The penalty parameter 'lam'
        will be overriden by with penalty selected by ms_estimator.
        If None, this will use QuicGraphicalLasso.

    graph :  An instance of a class with the method .create(n_features, alpha)
        that returns (cov, prec, adj).
        graph.create() will be used to draw a new graph instance in each trial.
        default: ErdosRenyiGraph()

    n_samples_grid : int (default=10) or array of floats
        Grid points for choosing number of samples.
        If integer, defines a linear grid (5, 200)
        Else uses array as grid.

    alpha_grid : int (default=5) or array of floats
        Grid points used in making new graphs via graph.create().
        If integer, defines a logarithmic grid (0.14, 0.4)
        Else, uses array as grid.

    metrics : dict of functions: scalar = func(prec, prec_estimate)
        The key for each function will be used for reporting results.
        default: {'frobenius': error_fro}

    verbose : bool (default=False)
        Print out progress information.

    n_jobs: int (optional)
        number of jobs to run in parallel (default 1).

    sc: sparkContext (optional)
        If a sparkContext object is provided, n_jobs will be ignore and the
        work will be parallelized via spark.

    seed : np.random.RandomState starting seed. (default=2)


    Attributes
    ----------
    grid_ : array of size (n_samples_grid, ) or n_samples_grid
        Array of n_samples / n_features ratios.

    alphas_ : array of size (alpha_grid, ) or alpha_grid
        Array of alphas used to generate graphs.

    precision_nnz_ : array of size (len(alphas_), )
        The sparsity of each test graph precision.

    results_ : dict of matrices of size (len(alphas_), len(grid_))
        Each key corresponds to a function from metrics.
    """

    def __init__(
        self,
        n_features=50,
        n_trials=100,
        ms_estimator=None,
        mc_estimator=None,
        graph=None,
        n_samples_grid=10,
        alpha_grid=5,
        metrics={"frobenius": error_fro},
        verbose=False,
        n_jobs=1,
        sc=None,
        seed=2,
    ):
        self.n_features = n_features
        self.n_trials = n_trials
        self.ms_estimator = ms_estimator
        self.mc_estimator = mc_estimator
        self.graph = graph
        self.n_samples_grid = n_samples_grid
        self.alpha_grid = alpha_grid
        self.metrics = metrics
        self.verbose = verbose
        self.n_jobs = n_jobs
        self.sc = sc
        self.seed = seed
        self.prng = np.random.RandomState(seed)

        if self.graph is None:
            self.graph = ErdosRenyiGraph()

        if self.ms_estimator is None:
            self.ms_estimator = QuicGraphicalLassoCV()

        if self.mc_estimator is None:
            self.mc_estimator = QuicGraphicalLasso(
                lam=0.5, mode="default", init_method="corrcoef"
            )

        if isinstance(self.n_samples_grid, int):
            self.grid_ = np.linspace(5, 200, self.n_samples_grid)
        else:
            self.grid_ = self.n_samples_grid

        if isinstance(self.alpha_grid, int):
            self.alphas_ = np.logspace(np.log(0.15), np.log10(0.4), self.alpha_grid)
        else:
            self.alphas_ = self.alpha_grid

        self.is_fitted = False
        self.results_ = None
        self.precision_nnz_ = None

    def fit(self, X=None, y=None):
        n_alphas = len(self.alphas_)
        n_grids = len(self.grid_)

        self.precision_nnz_ = []
        self.results_ = {k: np.zeros((n_alphas, n_grids)) for k in self.metrics}

        # build an indexed set (or generator) of grid points
        param_grid = [(a, g) for a in self.alphas_ for g in self.grid_]
        indexed_param_grid = list(zip(range(len(param_grid)), param_grid))

        ms_fit = partial(
            _ms_fit,
            estimator=self.ms_estimator,
            n_features=self.n_features,
            graph=self.graph,
            prng=self.prng,
        )

        if self.verbose:
            print("Getting parameters via model selection...")

        if self.sc is not None:
            ms_results = _spark_map(ms_fit, indexed_param_grid, self.sc, self.seed)
        else:
            ms_results = _cpu_map(ms_fit, indexed_param_grid, self.n_jobs, self.verbose)

        # ensure results are ordered
        ms_results = sorted(ms_results, key=lambda r: r[0])

        # track nnz of graph precision
        self.precision_nnz_ = [
            np.count_nonzero(graph[1].flat) for _, (graph, _, _) in ms_results
        ]

        # build param grid for mc trials
        # following results in an grid where nn indexes each trial of each
        # param grid:
        #  (0, graph_0, lam_0, n_samples_), (1, graph_0, lam_0, n_samples_0),..
        trial_param_grid = [
            (nn, graph, lam, n_samples)
            for _, (graph, lam, n_samples) in ms_results
            for nn in range(self.n_trials)
        ]
        indexed_trial_param_grid = list(
            zip(range(len(trial_param_grid)), trial_param_grid)
        )

        mc_fit = partial(_mc_fit, estimator=self.mc_estimator, metrics=self.metrics)

        if self.verbose:
            print("Fitting MC trials...")

        if self.sc is not None:
            mc_results = _spark_map(
                mc_fit, indexed_trial_param_grid, self.sc, self.seed + len(param_grid)
            )
        else:
            mc_results = _cpu_map(
                partial(mc_fit, prng=self.prng),
                indexed_trial_param_grid,
                self.n_jobs,
                self.verbose,
            )

        # ensure results are ordered correctly
        mc_results = sorted(mc_results, key=lambda r: r[0])

        # strip mc_results
        mc_results = [r for _, r in mc_results]

        # reduce
        param_grid_matrix_index = [
            (a, g) for a in range(len(self.alphas_)) for g in range(len(self.grid_))
        ]
        for param_index in range(len(param_grid)):
            trial_start = param_index * self.n_trials
            trials = mc_results[trial_start : trial_start + self.n_trials]
            aidx, gidx = param_grid_matrix_index[param_index]
            for key in self.metrics:
                results_by_key = np.array([t[key] for t in trials])
                self.results_[key][aidx, gidx] = (
                    1.0 * np.sum(results_by_key) / self.n_trials
                )

        if self.verbose:
            for key in self.metrics:
                print("Results for {}: {}".format(key, self.results_[key][aidx, :]))

        self.is_fitted = True
        return self



from __future__ import absolute_import
from .inverse_covariance import InverseCovarianceEstimator
from .quic_graph_lasso import (
    quic,
    QuicGraphLasso,
    QuicGraphLassoCV,
    QuicGraphLassoEBIC,
    QuicGraphicalLasso,
    QuicGraphicalLassoCV,
    QuicGraphicalLassoEBIC,
)
from .metrics import log_likelihood, kl_loss, quadratic_loss, ebic
from .rank_correlation import spearman_correlation, kendalltau_correlation
from .model_average import ModelAverage
from .adaptive_graph_lasso import AdaptiveGraphLasso, AdaptiveGraphicalLasso

__all__ = [
    "InverseCovarianceEstimator",
    "quic",
    "QuicGraphLasso",
    "QuicGraphLassoCV",
    "QuicGraphLassoEBIC",
    "QuicGraphicalLasso",
    "QuicGraphicalLassoCV",
    "QuicGraphicalLassoEBIC",
    "log_likelihood",
    "kl_loss",
    "quadratic_loss",
    "ebic",
    "spearman_correlation",
    "kendalltau_correlation",
    "ModelAverage",
    "AdaptiveGraphLasso",
    "AdaptiveGraphicalLasso",
]



from __future__ import absolute_import

from .graphs import Graph, lattice


class LatticeGraph(Graph):
    """Returns the adjacency matrix for a lattice/banded network via .create().

    The graph can be made fully connected using chaining assumption when
    chain_blocks=True (default).

    Parameters
    -----------
    random sign : bool (default=False)
        Randomly modulate each entry by 1 or -1 with probability of 1/2.

    low : float (0, 1) (default=0.3)
        Lower bound for np.random.RandomState.uniform before normalization.

    high : float (0, 1) > low (default=0.7)
        Upper bound for np.random.RandomState.uniform before normalization.

    n_blocks : int (default=2)
        Number of blocks.  Returned matrix will be square with
        shape n_block_features * n_blocks.

    chain_blocks : bool (default=True)
        Apply random lattice structure to chain blocks together.

    seed : int
        Seed for np.random.RandomState seed. (default=1)
    """

    def __init__(self, random_sign=False, low=0.7, high=0.7, **kwargs):
        self.random_sign = random_sign
        self.low = low
        self.high = high
        super(LatticeGraph, self).__init__(**kwargs)

    def prototype_adjacency(self, n_block_features, alpha):
        """Build a new graph.

        Doc for ".create(n_features, alpha)"

        Parameters
        -----------
        n_features : int

        alpha : float (0,1)
            The complexity / sparsity factor.

            Each graph will have a minimum of

                n_blocks * ceil(alpha * n_block_features)

                where

                n_block_features = floor(n_features / self.n_blocks)

            edges and exactly this amount if chain_blocks=False.

        Returns
        -----------
        (n_features, n_features) matrices: covariance, precision, adjacency
        """
        return lattice(
            self.prng,
            n_block_features,
            alpha,
            random_sign=self.random_sign,
            low=self.low,
            high=self.high,
        )



import numpy
from distutils.core import setup
from distutils.extension import Extension
from Cython.Distutils import build_ext
import platform


if platform.system() == "Darwin":
    extra_compile_args = ["-I/System/Library/Frameworks/vecLib.framework/Headers"]
    if "ppc" in platform.machine():
        extra_compile_args.append("-faltivec")

    extra_link_args = ["-Wl,-framework", "-Wl,Accelerate"]
    include_dirs = [numpy.get_include()]

else:
    include_dirs = [numpy.get_include(), "/usr/local/include"]
    extra_compile_args = ["-msse2", "-O2", "-fPIC", "-w"]
    extra_link_args = ["-llapack"]

ext_module = Extension(
    name="pyquic",
    sources=["QUIC.C", "pyquic.pyx"],
    include_dirs=include_dirs,
    extra_compile_args=extra_compile_args,
    extra_link_args=extra_link_args,
    language="c++",
)

if __name__ == "__main__":
    setup(name="pyquic", cmdclass={"build_ext": build_ext}, ext_modules=[ext_module])



from __future__ import print_function
import sys
from setuptools import setup
from distutils.extension import Extension
from Cython.Build import cythonize
import platform


try:
    import numpy  # NOQA
except ImportError:
    print('numpy is required during installation')
    sys.exit(1)

try:
    import scipy  # NOQA
except ImportError:
    print('scipy is required during installation')
    sys.exit(1)

try:
    import Cython  # NOQA
except ImportError:
    print('Cython is required during installation')
    sys.exit(1)

import numpy as np

if platform.system() == 'Darwin':
    extra_compile_args = [
        '-I/System/Library/Frameworks/vecLib.framework/Headers'
    ]
    if 'ppc' in platform.machine():
        extra_compile_args.append('-faltivec')

    extra_link_args = ["-Wl,-framework", "-Wl,Accelerate"]
    include_dirs = [np.get_include()]

else:
    include_dirs = [np.get_include(), "/usr/local/include"]
    extra_compile_args = ['-msse2', '-O2', '-fPIC', '-w']
    extra_link_args = ["-llapack"]


# pyquic extension
# --> inverse_covariance.pyquic.pyquic (contains func quic)
ext_module = Extension(
    name="pyquic.pyquic",  # note: we ext_package= flag in setup()
    sources=[
        "inverse_covariance/pyquic/QUIC.C",
        "inverse_covariance/pyquic/pyquic.pyx"],
    include_dirs=include_dirs,
    extra_compile_args=extra_compile_args,
    extra_link_args=extra_link_args,
    language="c++"
)


with open('requirements.txt') as f:
    INSTALL_REQUIRES = [l.strip() for l in f.readlines() if l]


setup(
    name='skggm2',
    version='0.2.9',
    description='Gaussian graphical models for scikit-learn.',
    author='Jason Laska and Manjari Narayan',
    license='MIT',
    packages=[
        'inverse_covariance',
        'inverse_covariance.profiling',
        'inverse_covariance.pyquic'],
    install_requires=INSTALL_REQUIRES,
    url='https://github.com/skggm/skggm',
    author_email='jlaska@gmail.com',
    ext_package='inverse_covariance',
    ext_modules=cythonize(ext_module),
)



"""Various utilities for Gaussian graphical models."""
import sys
import numpy as np
from sklearn.utils._testing import assert_array_equal
from matplotlib import pyplot as plt
import seaborn  # NOQA

plt.ion()


def r_input(val):
    if sys.version_info[0] >= 3:
        return eval(input(val))

    return raw_input(val)  # NOQA


def _check_path(in_path):
    path = sorted(set(in_path), reverse=True)
    assert_array_equal(path, in_path)


def trace_plot(precisions, path, n_edges=20, ground_truth=None, edges=[]):
    """Plot the change in precision (or covariance) coefficients as a function
    of changing lambda and l1-norm.  Always ignores diagonals.

    Parameters
    -----------
    precisions : array of len(path) 2D ndarray, shape (n_features, n_features)
        This is either precision_ or covariance_ from an InverseCovariance
        estimator in path mode, or a list of results for individual runs of
        the GraphLasso.

    path :  array of floats (descending)
        This is path of lambdas explored.

    n_edges :  int (default=20)
        Max number of edges to plot for each precision matrix along the path.
        Only plots the maximum magnitude values (evaluating the last precision
        matrix).

    ground_truth : 2D ndarray, shape (n_features, n_features) (default=None)
        If not None, plot the top n_edges/2 false positive and top n_edges/2
        false negative indices when compared to ground_truth.

    edges : list (default=[])
        If not empty, use edges to determine which indicies of each precision
        matrix to track.  Should be arranged to index precisions[0].flat.

        If non-empty, n_edges and ground_truth will be ignored.
    """
    _check_path(path)
    assert len(path) == len(precisions)
    assert len(precisions) > 0

    path = np.array(path)
    dim, _ = precisions[0].shape

    # determine which indices to track
    if not edges:
        base_precision = np.copy(precisions[-1])
        base_precision[np.triu_indices(base_precision.shape[0])] = 0

        if ground_truth is None:
            # top n_edges strongest coefficients
            edges = np.argsort(np.abs(base_precision.flat))[::-1][:n_edges]
        else:
            # top n_edges/2 false positives and negatives compared to truth
            assert ground_truth.shape == precisions[0].shape
            masked_gt = np.copy(ground_truth)
            masked_gt[np.triu_indices(ground_truth.shape[0])] = 0

            intersection = np.intersect1d(
                np.nonzero(base_precision.flat)[0], np.nonzero(masked_gt.flat)[0]
            )

            # false positives
            fp_precision = np.copy(base_precision)
            fp_precision.flat[intersection] = 0
            fp_edges = np.argsort(np.abs(fp_precision.flat))[::-1][: n_edges / 2]

            # false negatives
            fn_precision = np.copy(masked_gt)
            fn_precision.flat[intersection] = 0
            fn_edges = np.argsort(np.abs(fn_precision.flat))[::-1][: n_edges / 2]

            edges = list(fp_edges) + list(fn_edges)

    assert len(edges) < len(precisions[0].flat)
    assert np.max(edges) < len(precisions[0].flat)
    assert np.min(edges) >= 0

    # reshape data a bit:
    # flatten each matrix into a column (so that coeffs are examples)
    # compute l1-norm of each column
    l1_norms = []
    coeffs = np.zeros((dim**2, len(precisions)))
    for ridx, result in enumerate(precisions):
        coeffs[edges, ridx] = result.flat[edges]
        l1_norms.append(np.linalg.norm(coeffs[:, ridx]))

    # remove any zero rows
    coeffs = coeffs[np.linalg.norm(coeffs, axis=1) > 1e-10, :]

    plt.figure()

    # show coefficients as a function of lambda
    plt.subplot(1, 2, 1)
    for result in precisions:
        plt.plot(l1_norms, coeffs.T, lw=1)

    plt.xlim([np.min(l1_norms), np.max(l1_norms)])
    plt.ylabel("Coefficients")
    plt.xlabel("l1 Norm")

    # show coefficients as a function of lambda
    log_path = np.log(path)
    plt.subplot(1, 2, 2)
    for result in precisions:
        plt.plot(log_path, coeffs.T, lw=1)

    plt.xlim([np.min(log_path), np.max(log_path)])
    plt.ylabel("Coefficients")
    plt.xlabel("log-Lambda")

    plt.show()
    r_input("Press any key to continue.")



from __future__ import absolute_import

import numpy as np
from sklearn.utils import check_array, as_float_array, deprecated
from sklearn.base import BaseEstimator

from . import QuicGraphicalLasso, QuicGraphicalLassoCV, InverseCovarianceEstimator


class AdaptiveGraphicalLasso(BaseEstimator):
    """
    Two-stage adaptive estimator.

    a) Compute the first estimate with an estimator of your choice
       (self.estimator).  This estimate must produce a single precision_
       output.

    b) The resulting coefficients are used to generate adaptive weights and
       the QuicGraphicalLasso is refit with these weights.

    See:
        "High dimensional covariance estimation based on Gaussian graphical
        models"
        S. Zhou, P. R{\"u}htimann, M. Xu, and P. B{\"u}hlmann

        "Relaxed Lasso"
        N. Meinshausen, December 2006.

    Parameters
    -----------
    estimator : GraphicalLasso instance with model selection
                (default=QuicGraphicalLassoCV())
        After being fit, estimator.precision_ must either be a matrix.

    method : one of 'binary', 'inverse_squared', 'inverse' (default='binary')
        binary: non-zero where estimator was zero, 1 else
                (gelato, "relaxed lasso")
        inverse_squared: 1/|coefficient|^2  (glasso)
        inverse: 1/|coefficient|

    Attributes
    ----------
    estimator_ : QuicGraphicalLasso instance
        The final estimator refit with adaptive weights.

    lam_ : 2D ndarray, shape (n_features, n_features)
        Adaptive weight matrix generated.
    """

    def __init__(self, estimator=None, method="binary"):
        self.estimator = estimator
        self.method = method

    def _binary_weights(self, estimator):
        n_features, _ = estimator.precision_.shape
        lam = np.zeros((n_features, n_features))
        lam[estimator.precision_ == 0] = 1
        lam[np.diag_indices(n_features)] = 0
        return lam

    def _inverse_squared_weights(self, estimator):
        n_features, _ = estimator.precision_.shape
        lam = np.zeros((n_features, n_features))
        mask = estimator.precision_ != 0
        lam[mask] = 1.0 / (np.abs(estimator.precision_[mask]) ** 2)
        mask_0 = estimator.precision_ == 0
        lam[mask_0] = np.max(lam[mask].flat)  # non-zero in appropriate range
        lam[np.diag_indices(n_features)] = 0
        return lam

    def _inverse_weights(self, estimator):
        n_features, _ = estimator.precision_.shape
        lam = np.zeros((n_features, n_features))
        mask = estimator.precision_ != 0
        lam[mask] = 1.0 / np.abs(estimator.precision_[mask])
        mask_0 = estimator.precision_ == 0
        lam[mask_0] = np.max(lam[mask].flat)  # non-zero in appropriate range
        lam[np.diag_indices(n_features)] = 0
        return lam

    def fit(self, X, y=None):
        """Estimate the precision using an adaptive maximum likelihood estimator.
        Parameters
        ----------
        X : ndarray, shape (n_samples, n_features)
            Data from which to compute the proportion matrix.
        """
        # default to QuicGraphicalLassoCV
        estimator = self.estimator or QuicGraphicalLassoCV()

        self.lam_ = None
        self.estimator_ = None

        X = check_array(X, ensure_min_features=2, estimator=self)
        X = as_float_array(X, copy=False, force_all_finite=False)

        n_samples_, n_features_ = X.shape

        # perform first estimate
        estimator.fit(X)

        if self.method == "binary":
            # generate weights
            self.lam_ = self._binary_weights(estimator)

            # perform second step adaptive estimate
            self.estimator_ = QuicGraphicalLasso(
                lam=self.lam_ * estimator.lam_,
                mode="default",
                init_method="cov",
                auto_scale=False,
            )
            self.estimator_.fit(X)

        elif self.method == "inverse_squared":
            self.lam_ = self._inverse_squared_weights(estimator)

            # perform second step adaptive estimate
            self.estimator_ = QuicGraphicalLassoCV(
                lam=self.lam_ * self.estimator.lam_, auto_scale=False
            )
            self.estimator_.fit(X)

        elif self.method == "inverse":
            self.lam_ = self._inverse_weights(estimator)

            # perform second step adaptive estimate
            self.estimator_ = QuicGraphicalLassoCV(
                lam=self.lam_ * estimator.lam_, auto_scale=False
            )
            self.estimator_.fit(X)

        else:
            raise NotImplementedError(
                (
                    "Only method='binary', 'inverse_squared', or",
                    "'inverse' have been implemented.",
                )
            )

        self.is_fitted_ = True
        self.n_features_in_ = X.shape[1]
        return self


@deprecated(
    "The class AdaptiveGraphLasso is deprecated "
    "Use class AdaptiveGraphicalLasso instead."
)
class AdaptiveGraphLasso(AdaptiveGraphicalLasso):
    pass



import numpy as np
import scipy as sp


def lattice(prng, n_features, alpha, random_sign=False, low=0.3, high=0.7):
    """Returns the adjacency matrix for a lattice network.

    The resulting network is a Toeplitz matrix with random values summing
    between -1 and 1 and zeros along the diagonal.

    The range of the values can be controlled via the parameters low and high.
    If random_sign is false, all entries will be negative, otherwise their sign
    will be modulated at random with probability 1/2.

    Each row has maximum edges of np.ceil(alpha * n_features).

    Parameters
    -----------
    n_features : int

    alpha : float (0, 1)
        The complexity / sparsity factor.

    random sign : bool (default=False)
        Randomly modulate each entry by 1 or -1 with probability of 1/2.

    low : float (0, 1) (default=0.3)
        Lower bound for np.random.RandomState.uniform before normalization.

    high : float (0, 1) > low (default=0.7)
        Upper bound for np.random.RandomState.uniform before normalization.
    """
    degree = int(1 + np.round(alpha * n_features / 2.0))

    if random_sign:
        sign_row = -1.0 * np.ones(degree) + 2 * (
            prng.uniform(low=0, high=1, size=degree) > 0.5
        )
    else:
        sign_row = -1.0 * np.ones(degree)

    # in the *very unlikely* event that we draw a bad row that sums to zero
    # (which is only possible when random_sign=True), we try again up to
    # MAX_ATTEMPTS=5 times.  If we are still unable to draw a good set of
    # values something is probably wrong and we raise.
    MAX_ATTEMPTS = 5
    attempt = 0
    row = np.zeros((n_features,))
    while np.sum(row) == 0 and attempt < MAX_ATTEMPTS:
        row = np.zeros((n_features,))
        row[1 : 1 + degree] = sign_row * prng.uniform(low=low, high=high, size=degree)
        attempt += 1

    if np.sum(row) == 0:
        raise Exception("InvalidLattice", "Rows sum to 0.")
        return

    # sum-normalize and keep signs
    row /= np.abs(np.sum(row))

    return sp.linalg.toeplitz(c=row, r=row)


def blocks(prng, block, n_blocks=2, chain_blocks=True):
    """Replicates `block` matrix n_blocks times diagonally to create a
    square matrix of size n_features = block.size[0] * n_blocks and with zeros
    along the diagonal.

    The graph can be made fully connected using chaining assumption when
    chain_blocks=True (default).

    This utility can be used to generate cluster networks or banded lattice n
    networks, among others.

    Parameters
    -----------
    block : 2D array (n_block_features, n_block_features)
        Prototype block adjacency matrix.

    n_blocks : int (default=2)
        Number of blocks.  Returned matrix will be square with
        shape n_block_features * n_blocks.

    chain_blocks : bool (default=True)
        Apply random lattice structure to chain blocks together.
    """
    n_block_features, _ = block.shape
    n_features = n_block_features * n_blocks
    adjacency = np.zeros((n_features, n_features))

    dep_groups = np.eye(n_blocks)
    if chain_blocks:
        chain_alpha = np.round(0.01 + 0.5 / n_blocks, 2)
        chain_groups = lattice(prng, n_blocks, chain_alpha, random_sign=False)
        chain_groups *= -0.1
        dep_groups += chain_groups

    adjacency = np.kron(dep_groups, block)
    adjacency[np.where(np.eye(n_features))] = 0
    return adjacency


def _to_diagonally_dominant(mat):
    """Make matrix unweighted diagonally dominant using the Laplacian."""
    mat += np.diag(np.sum(mat != 0, axis=1) + 0.01)
    return mat


def _to_diagonally_dominant_weighted(mat):
    """Make matrix weighted diagonally dominant using the Laplacian."""
    mat += np.diag(np.sum(np.abs(mat), axis=1) + 0.01)
    return mat


def _rescale_to_unit_diagonals(mat):
    """Rescale matrix to have unit diagonals.

    Note: Call only after diagonal dominance is ensured.
    """
    d = np.sqrt(np.diag(mat))
    mat /= d
    mat /= d[:, np.newaxis]
    return mat


class Graph(object):
    """Base class that returns the adjacency matrix for a network via
    the .create() method.

    Parameters
    -----------
    n_blocks : int (default=2)
        Number of blocks.  Returned matrix will be square with
        shape n_block_features * n_blocks.

    chain_blocks : bool (default=True)
        Apply random lattice structure to chain blocks together.

    seed : int
        Seed for np.random.RandomState seed. (default=1)
    """

    def __init__(self, n_blocks=2, chain_blocks=True, seed=1):
        self.n_blocks = n_blocks
        self.chain_blocks = chain_blocks
        self.seed = seed
        self.prng = np.random.RandomState(self.seed)

        if n_blocks == 1 and chain_blocks:
            raise ValueError("Requires chain_blocks=False when n_blocks=1.")
            return

    def to_precision(self, adjacency, weighted=True, rescale=True):
        if weighted:
            dd_adj = _to_diagonally_dominant_weighted(adjacency)
        else:
            dd_adj = _to_diagonally_dominant(adjacency)

        if rescale:
            return _rescale_to_unit_diagonals(dd_adj)

        return dd_adj

    def to_covariance(self, precision, rescale=True):
        covariance = np.linalg.inv(precision)

        if rescale:
            return _rescale_to_unit_diagonals(covariance)

        return covariance

    def prototype_adjacency(self, n_block_features, alpha):
        """Override this method with a custom base graph."""
        pass

    def create(self, n_features, alpha):
        """Build a new graph with block structure.

        Parameters
        -----------
        n_features : int

        alpha : float (0,1)
            The complexity / sparsity factor for each graph type.

        Returns
        -----------
        (n_features, n_features) matrices: covariance, precision, adjacency
        """
        n_block_features = int(np.floor(1.0 * n_features / self.n_blocks))
        if n_block_features * self.n_blocks != n_features:
            raise ValueError(
                (
                    "Error: n_features {} not divisible by n_blocks {}."
                    "Use n_features = n_blocks * int"
                ).format(n_features, self.n_blocks)
            )
            return

        block_adj = self.prototype_adjacency(n_block_features, alpha)
        adjacency = blocks(
            self.prng, block_adj, n_blocks=self.n_blocks, chain_blocks=self.chain_blocks
        )

        precision = self.to_precision(adjacency)
        covariance = self.to_covariance(precision)
        return covariance, precision, adjacency
